/*  
    Copyright (C) 2004 Julia Handl
    Email: Julia.Handl@gmx.de

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/


#ifndef DATABIN_JH_2003
#define DATABIN_JH_2003

#include "conf.h"
#include "databin.h"
#include <fstream>
#include "math.h"
#include "random.h"
#include "tmatrix.h"


template <class BINTYPE> class databin;
template <class DBINTYPE> class docbin;

extern long idum;



/***************************************************

       class data

***************************************************/


#define NRANSI
#define MAXBIT 30
#define MAXDIM 6
static int iminarg1,iminarg2;
#define IMIN(a,b) (iminarg1=(a),iminarg2=(b),(iminarg1) < (iminarg2) ? (iminarg1) : (iminarg2))






template <class DATATYPE>
class data {
 
    friend class databin<DATATYPE>;

 private:
    DATATYPE * vector;
  
protected:
    conf * par;

   
 public:
    int cluster;
    int color;
    ~data();


    /* Default constructor */
    data(conf * c);

    /* Constructor for a data item that is initialized to the vector <d> */
    data(conf * c, DATATYPE * d);

    /* Constructor for a data item that is initialized to the vector <d> and 
       assigned cluster number <cluster> and color <color> */
    data(conf * c, DATATYPE * d, int color, int cluster);

    /* Lenght of the data vector */
    const int length();

    DATATYPE square(DATATYPE x);

    /* Write-Read access to individual components */
    DATATYPE &operator[](const int i);

    /* Distance computation between two data items */
    const DATATYPE distanceto(data<DATATYPE> & d);

    /* Addition of two data vectors */
    void add(data<DATATYPE> & d);

    /* Division of a data vector by <i> */
    void div(int i);

    /* Data vector is set to <d> */
    void set(data<DATATYPE> & d);

  /* Data vector is set to <d> */
    void set(DATATYPE * d);


  
};




/***************************************************

       class databin

***************************************************/


template <class BINTYPE>
class databin {

 protected:
    /* pointer to current parameter settings */
    conf * par; 

    /* array of pointer to data objects */ 
    data<BINTYPE> ** bin;   

   
    int * color;

    /* precomputed disimilarity matrix */
    tmatrix<BINTYPE> * distancematrix;

     
  
    
 public:

    BINTYPE * minvalue;
    BINTYPE * maxvalue;

    databin(conf * c, char * name, int size, int dim, int clust);


    databin(conf * c,BINTYPE ** t);

  /* Destructor */
  ~databin();

  /* Distance computation between two data items in the collection */
  const BINTYPE d(const int index1, const int index2);

 /* Direct access to precomputed dissimilarity matrix */
  const BINTYPE precomputed_d(const int index1, const int index2);

 /* Permute order of data items */
  void permutate();

  void uniformprescription();

  /* Write-Read access to an individual data item in the collection */
  data<BINTYPE> & operator[](const int i);

 /* Check whether this class label already exists */
  int find(char * templabel, char ** classlabel, int & labelctr);


  double max(double x, double y);
  double min(double x, double y);
 

};




/***************************************************

       class data - function definitions

***************************************************/


// access to vector length
template <class DATATYPE>const int data <DATATYPE>::length() {
    return par->bindim;
}


 // default constructor
template <class DATATYPE>data <DATATYPE>::data(conf * c)  {
    par = c;
    vector = new DATATYPE[par->bindim];
  
    
    for (int i=0; i<par->bindim; i++) {
	vector[i] = 0;
    }
}

 // constructor if data vector is provided
template <class DATATYPE>data <DATATYPE>:: data(conf * c, DATATYPE * dat, int col, int cl ) {
    par = c;
    vector = new DATATYPE[par->bindim];
    color = col;
    cluster = cl;
      
    for (int i=0; i<par->bindim; i++) {
	vector[i] = dat[i];
	//	cerr << vector[i] << " ";
    }
    //  cerr << endl;

}


 // constructor if data vector is provided
template <class DATATYPE>data <DATATYPE>:: data(conf * c, DATATYPE * dat) {
    par = c;
    vector = new DATATYPE[par->bindim];
   
      
    for (int i=0; i<par->bindim; i++) {
	vector[i] = dat[i];
    }
}


// destructor
template <class DATATYPE>data <DATATYPE>::~data() {
    delete [] vector;
}


// square function
template <class DATATYPE> DATATYPE data <DATATYPE>::square(DATATYPE x) {
    return x*x;
}

// write-read access to coordinates of dat vector 
template <class DATATYPE> DATATYPE &data <DATATYPE>::operator[](const int i) {
    return vector[i];
}

// distance function defined between data vectors   
template <class DATATYPE>const DATATYPE data <DATATYPE>::distanceto(data<DATATYPE> & dd) {
  

    data & d = (data &)dd;
    DATATYPE result = 0.0;

    for (int i=0; i<par->bindim; i++) {
	result += square(d.vector[i] - vector[i]);
    }
    
    return sqrt(result);



}



// addition of data vectors   
template <class DATATYPE>void data <DATATYPE>::add(data<DATATYPE> & d) {
    for (int i=0; i<par->bindim; i++) {
	vector[i] += d.vector[i];

    }
 
}

template <class DATATYPE> void data<DATATYPE>::set(data<DATATYPE> & d) {
      for (int i=0; i<par->bindim; i++) {
	  vector[i] = d.vector[i];
      }
}

template <class DATATYPE> void data<DATATYPE>::set(DATATYPE * d) {
      for (int i=0; i<par->bindim; i++) {
	  vector[i] = d[i];
      }
}



// division of a data vector by an integer   
template <class DATATYPE>void data <DATATYPE>::div(int divisor) {
    for (int i=0; i<par->bindim; i++) {
	vector[i] /= double(divisor);

    }
}



/***************************************************

       class databin - function definitions

***************************************************/



// destructor
template <class BINTYPE>databin <BINTYPE>::~databin() {


  //  cerr << par->binsize << endl;
  for (int i=0; i<par->binsize; i++) {
	delete bin[i];
    }
  delete [] bin;
  delete [] minvalue;
  delete [] maxvalue;
  if (color != NULL) delete [] color;
   delete distancematrix;
}



// distance function between the bin's data items
template <class BINTYPE>inline const BINTYPE databin <BINTYPE>::d(const int index1, const int index2) {
    return bin[index1]->distanceto(*bin[index2]);
}

template <class BINTYPE> const BINTYPE databin <BINTYPE>::precomputed_d(const int index1, const int index2) {
    return (*distancematrix)(index1,index2);
}


// write-read access to data items
template <class BINTYPE>inline data<BINTYPE> & databin <BINTYPE>::operator[](const int i) {
    return *bin[i];
}

 


 



// constructor for class databin if a real data set is used (identified by name) 
template <class BINTYPE>databin <BINTYPE>::databin(conf * c, char * name, int size, int dim, int clust) {

    int CL = clust;
    clust = 100; 

    ifstream input(name);
    if (input.bad()) {
      cerr << "Error while trying to open file " << name << endl;
      exit(0);
    }
   
    int labelctr = 0;
    char ** classlabels = new char *[size];
    for (int i=0; i<size; i++) {
      classlabels[i] = new char[100];
    }
    par = c;
    par->bindim = dim;
    par->binsize = size;
    par->kclusters = CL;
    par->imax_som = 1;
    par->jmax_som = par->kclusters;
    par->num_cluster = clust;
    if (par->size_cluster != NULL) 
      par->size_cluster = new int[par->num_cluster];
    
    color = new int[par->binsize];
    minvalue = new BINTYPE[par->bindim];
    maxvalue = new BINTYPE[par->bindim];

    for (int i=0; i<par->num_cluster; i++) {
      par->size_cluster[i] = 0;
    }


    bin = new data<USED_DATA_TYPE>*[par->binsize];
   

    USED_DATA_TYPE ** temp = new USED_DATA_TYPE*[par->binsize];
    for (int i=0; i<par->binsize; i++) {
	temp[i] = new USED_DATA_TYPE[par->bindim];
    }

    char templabel[100];
    par->num_cluster = 0;
    for (int i=0; i<par->binsize; i++) {
	for (int j=0; j<par->bindim; j++) {
	    input >> temp[i][j];
	  
	    if (i == 0) minvalue[j] = temp[i][j];
	    else minvalue[j] = min(minvalue[j], temp[i][j]);
	    if (i == 0) maxvalue[j] = temp[i][j];
	    else maxvalue[j] = max(maxvalue[j], temp[i][j]);
	}
	input >> templabel;


	color[i] = find(templabel, classlabels, labelctr);
	par->size_cluster[color[i]]++;
	par->num_cluster = (int)max(par->num_cluster, color[i]+1);
    }


    for (int i=0; i<par->binsize; i++) {
      bin[i] = new data<USED_DATA_TYPE>(par, temp[i], color[i]+1, color[i]);
    }


    distancematrix = new tmatrix<USED_DATA_TYPE>(par->binsize);
    
  
    // compute distances 
  
    for (int i=0; i<par->binsize; i++) {
	for (int j=0; j<i; j++) {
	    (*distancematrix)(i,j) = bin[i]->distanceto(*(bin[j]));
	}
    }

    
    for (int i=0; i<par->binsize; i++) {
	delete [] temp[i];
    }
    delete [] temp;



    for (int i=0; i<par->binsize; i++) {
      delete [] classlabels[i];
    }
    delete [] classlabels;

   

    par->kclusters = CL;
 

}



template <class BINTYPE>databin <BINTYPE>::databin(conf * c,BINTYPE ** t) {
    
    par = c;

    par->kclusters = par->num_cluster;
    par->imax_som = 1;
    par->jmax_som = par->kclusters;
   

    maxvalue = new USED_DATA_TYPE[par->bindim];
    minvalue = new USED_DATA_TYPE[par->bindim];
    color = NULL;
  
    
    bin = new data<USED_DATA_TYPE>*[par->binsize];
    if (bin == NULL) {
	cerr << "Databin: Memory allocation failed" << endl;
	exit(0);
    }


    int ctr = 0;
    int color = 0;
    for (int k=0; k<par->bindim; k++) {
	maxvalue[k] = -100000000.0;
	minvalue[k] = 100000000.0;
    }
    for (int j=0; j<par->binsize; j++) { 

      bin[ctr] = new data<USED_DATA_TYPE>(par, t[ctr]);
      if (bin[ctr] == NULL) {
	cerr << "Databin: Memory allocation failed" << endl;
	exit(0);
      }
      for (int k=0; k<par->bindim; k++) {
	maxvalue[k] = max(maxvalue[k], (*(bin[ctr]))[k]);
	minvalue[k] = min(minvalue[k], (*(bin[ctr]))[k]);
	
      }
	   
      ctr++;
    }
	

    
    distancematrix = new tmatrix<BINTYPE>(par->binsize);
    if (distancematrix == NULL) {
	cerr << "Databin: Memory allocation failed" << endl;
	exit(0);
    }
    
    // compute distances 
    for (int i=0; i<par->binsize; i++) {
	for (int j=0; j<i; j++) {
	  (*distancematrix)(i,j) = bin[i]->distanceto(*(bin[j]));
	}
    }
    
}


template <class BINTYPE> void databin <BINTYPE>::uniformprescription() {

  // long idum = rand();

    USED_DATA_TYPE ** datamatrix = new USED_DATA_TYPE * [par->binsize];
    double ** point = new double * [par->binsize];
    for (int i=0; i<par->binsize; i++) {
	datamatrix[i] = new USED_DATA_TYPE[par->bindim];
	point[i] = new double[par->bindim];
    }

    
    ofstream out("control.data");
    for(int t=0;t<par->binsize;t++)
    {  
      for (int j=0; j<par->bindim; j++) {
	datamatrix[t][j] = minvalue[j] + ran0(&idum)*(maxvalue[j]-minvalue[j]);
	out << datamatrix[t][j] << " ";
      }
      out << endl;
    }


    for (int i=0; i<par->binsize; i++) {
      delete bin[i];
    }

  
    for (int i=0; i<par->binsize; i++) {
      bin[i] = new data<USED_DATA_TYPE>(par, (USED_DATA_TYPE *)datamatrix[i], color[i]+1, color[i]);
      if (bin[i] == NULL) {
	cerr << "Databin: Memory allocation failed" << endl;
	exit(0);
      }
    }

    delete distancematrix;
    distancematrix = new tmatrix<BINTYPE>(par->binsize);
    if (distancematrix == NULL) {
	cerr << "Databin: Memory allocation failed" << endl;
	exit(0);
    }
   
    
    // compute distances
    for (int i=0; i<par->binsize; i++) {
	for (int j=0; j<i; j++) {
	  (*distancematrix)(i,j) = bin[i]->distanceto(*(bin[j]));
	}
    }


  
 
    for (int i=0; i<par->binsize; i++) {
      delete [] datamatrix[i];
      delete [] point[i];
    }
    delete [] datamatrix;
    delete [] point;


}


template <class BINTYPE> double databin <BINTYPE>::max(double x, double y) {
  if (x > y) return x;
  else return y;
}


template <class BINTYPE> double databin <BINTYPE>::min(double x, double y) {
  if (x < y) return x;
  else return y;
}


template <class BINTYPE> int databin <BINTYPE>::find(char * templabel, char ** classlabel, int & labelctr) {
    for (int i=0; i<labelctr; i++) {
	if (strcmp(classlabel[i],templabel) == 0) {
	    return i;
	}
    }
    strcpy(classlabel[labelctr], templabel);
    labelctr++;
    //   cout << "New Label: " << templabel << endl;
    return labelctr-1;
}

template <class BINTYPE> void databin <BINTYPE>::permutate() {

    BINTYPE tempval;

    for (int i=0; i<par->binsize; i++) {
	int j = int(ran0(&idum)*(par->binsize));

	data<BINTYPE> * temp = bin[i];
	bin[i] = bin[j];
	bin[j] = temp;
	for (int k=0; k<par->binsize; k++) {
	    if ((k != i)  && (k != j)) {
		tempval = (*distancematrix)(i,k);
		(*distancematrix)(i,k) = (*distancematrix)(j,k);
		(*distancematrix)(j,k) = tempval;
	    }
	  
	}


    }
    
 
}


#endif


