/*  
    Copyright (C) 2004 Julia Handl
    Email: Julia.Handl@gmx.de

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#ifndef EVALUATION_JH_2003
#define EVALUATION_JH_2003

#include "conf.h"
#include "databin.h"
#include "clustering.h"
#include <iostream.h>


class evaluation {
    
    conf * par;
    databin<USED_DATA_TYPE> * bin;
    clustering * clust;

   

   
 public:


    evaluation(conf * c);

    ~evaluation();

    void init(databin<USED_DATA_TYPE> * b, clustering * cl);

    double square(double x);
    
    /* Internal measures */
    // Intra -cluster variance
    double variance();
    // Connectivity
    double connectivity(int ** nnlist, int knn);
    // Silhouette Width
    double silhouette();
    // Dunn Index
    double dunn_max(); 


    /* External measures */
    // F-Measure
    double fmeasure(int b);
    // F-Measure with correction
    double adjustedfmeasure(int b);
    // Rand Index (for use as unary measure)
    double randindex();
    // Adjusted Rand Index (for use as unary measure)
    double adjustedrandindex();
    // Rand Index (for use as binary measure)
    double randindex(clustering * c1, clustering * c2, int size);
    // Tibshirani's predictive power (for use as binary measure)
    double tibshirani(clustering * c1, clustering * c2, int size);

  
    double min(double x, double y);
    double max(double x, double y);
    double fac(int xstart, int xend);
    int cover(int x, int n);
   
};



#endif
