
/*  
    Copyright (C) 2004 Julia Handl
    Email: Julia.Handl@gmx.de

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/


#ifndef SOTA_JH_2003
#define SOTA_JH_2003

#include "conf.h"
#include "databin.h"
#include "clalg.h"
#include "clustering.h"
#include "evaluation.h"

class node {

public: 
  double var;      // intra-node variance
  float * weight;  // weight vector
  node * left;     // daugher 1
  node * right;    // daughter 2 
  node * anc;      // parent
  int ctr;         // number of members associated with this node
  double d;        // intra-node variance
  int id;          // cluster number 
  int * member;    // members associated with this node

  node(conf * par); 
  ~node();


 
};




class sota:public clalg {

 private:


  
 public:

  node * root;

    /* Constructor */
    sota(conf * par, databin<USED_DATA_TYPE> * bin, evaluation * e);

    /* Destructor */
    ~sota();

    /* Initialisation of weight vectors */
    void init();
    
    /* Training phase */
    void run();

    /* Classification phase */
    void classify(); 

    void calculate(conf * par, node * go, double * epsilon, int * id);
    
    void setdiv(conf * par, node * current);

    node * maxvar(conf * par, node * x, double * var);

    node * mindist(conf * par, node * xx, double * var, data<USED_DATA_TYPE> * stimulus, int t);

    node * cluster(conf * par, node * xx, int t);


    /* Others */
    double square(double x);

    double myabs(double x);

    void del(node * root);

    

};





#endif
