/*  
    Copyright (C) 2004 Julia Handl
    Email: Julia.Handl@gmx.de

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/


#include <iostream.h>
#include "conf.h"
#include "evaluation.h"
#include <time.h>


#ifdef KMEANS
   #include "kmeans.h"
#endif

#ifdef AGGLOM
  #include "agglom.h"
#endif

#ifdef SOM
  #include "som.h"
#endif

#ifdef SOTA
  #include "sota.h"
#endif



using namespace std;

double square(double x);
void fileoutput(char *, double *, int);
void output(clustering * cl, int size);
void nnlist(int ** nnlist, int ** weight);
void sort(int ** nn_arg, int ** weight);
int lt2(const void *a, const void *b);

conf * par;
databin<USED_DATA_TYPE> * bin;
int ** nn;
int ** w;
USED_DATA_TYPE * dtemp;
long idum = time(NULL);


int main(int argc, char ** argv) {
 

    if (argc != 5) {
	cerr << "Four parameters expected: Filename, Size, Dimension and ClusterNumber" << endl;
	exit(0);
	}
    conf c;    
    evaluation e(&c);


/***************************************************************

 Agglomerativ hierarchical clustering

***************************************************************/

 
#ifdef AGGLOM


  
    agglom * arun;
    
    char name1[1000];
    char name2[1000];

#ifdef SINGLELINK
    sprintf(name1, "%s.single.solution", argv[1]);
    sprintf(name2, "%s.single.control", argv[1]);
#else
    sprintf(name1, "%s.agglom.solution", argv[1]);
    sprintf(name2, "%s.agglom.control", argv[1]);
#endif

    ofstream out1(name1);
    ofstream out2(name2);

    int ** neighbours;
    int ** weight;
    int size = atoi(argv[2]);

    neighbours = new int * [size];
    for (int i=0; i<size; i++) {
      neighbours[i] = new int[size-1];
    }
    weight  = new int * [size];
    for (int i=0; i<size; i++) {
      weight[i] = new int[size];
    }


    // Range of considered number of clusters
    for (int k = 1; k<=25; k++) {

           
      c.evalctr = 0;
      for (c.evalctr=0; c.evalctr<c.evalnbr; c.evalctr++) {

	// Read in data
	databin<float> d(&c, argv[1], atoi(argv[2]), atoi(argv[3]), k);
	par = &c;
	bin = &d;


	// Run clustering algorithm
	arun = new agglom(&c, &d, &e);

#ifdef SINGLELINK
	arun->single = true;
#endif

	arun->init();
	arun->run();

	// Compute nearest neigbor list needed for connectivity
	nnlist(neighbours, weight);

	// Apply validation technique
	e.init(&d, arun->getclustering());

	out1 << par->kclusters << " " << e.connectivity(nn,10) << " " << e.variance() << " " << e.dunn_max() << " " << e.silhouette() << " " << e.fmeasure(1) << " " << e.adjustedfmeasure(1) << " " << e.randindex() << " " << e.adjustedrandindex() << endl;


	delete arun;


	// Generate random control data
	par = &c;
	bin->uniformprescription();


	// Run clustering algorithm
	arun = new agglom(&c, bin, &e);

#ifdef SINGLELINK
	arun->single = true;
#endif

	arun->init();
	arun->run();


	// Compute nearest neighbor list needed for connectivity
	nnlist(neighbours, weight);

	// Apply validation etchniques
	e.init(bin, arun->getclustering());
	
	out2 << par->kclusters << " " << e.connectivity(nn,10) << " " << e.variance() << " " <<  e.dunn_max() << " " << e.silhouette() << " " << e.fmeasure(1) << " " << e.adjustedfmeasure(1) << " " << e.randindex() << " " << e.adjustedrandindex() << endl;


	delete arun;

       }
     }
    for (int i=0; i<par->binsize; i++) {
      delete [] neighbours[i];
    }
    for (int i=0; i<par->binsize; i++) {
      delete [] weight[i];
    }
    delete [] neighbours;
    delete [] weight;

 
    
#endif




/***************************************************************

 K-Means

***************************************************************/

#ifdef KMEANS 

    char name1[1000];
    char name2[1000];

    sprintf(name1, "%s.kmeans.solution", argv[1]);
    sprintf(name2, "%s.kmeans.control", argv[1]);


    ofstream out1(name1);
    ofstream out2(name2);
    

    int ** neighbours;
    int ** weight;
    int size = atoi(argv[2]);
    neighbours = new int * [size];
    for (int i=0; i<size; i++) {
      neighbours[i] = new int[size];
    }
    weight  = new int * [size];
    for (int i=0; i<size; i++) {
      weight[i] = new int[size];
    }


    // Range of considered number of clusters
    for (int k = 1; k <= 25; k++) {

    kmeans * krun;

    c.evalctr = 0;

     for (c.evalctr=0; c.evalctr<c.evalnbr; c.evalctr++) {

        // Read in data 
 	databin<float> d(&c, argv[1], atoi(argv[2]), atoi(argv[3]), k);

	par = &c;
	bin = &d;

	// Run clustering algorithm
	krun = new kmeans(&c, &d, &e);
	krun->init();
	krun->run();

	// Compute the nearest neighbor list needed for connectivity
	nnlist(neighbours, weight);

	// Apply validation techniques
	e.init(&d, krun->getclustering());

	out1 << par->kclusters << " " << e.connectivity(nn,10) << " " << e.variance() << " " << e.dunn_max() << " " << e.silhouette() << " " << e.fmeasure(1) << " " << e.adjustedfmeasure(1) << " " << e.randindex() << " " << e.adjustedrandindex() << endl;


	delete krun;


	par = &c;
	bin->uniformprescription();
    
	krun = new kmeans(&c, bin, &e);
    
	krun->init();
	krun->run();

    
	// Compute the nearest neighbor list needed for connectivity
	nnlist(neighbours, weight);
    

	// Apply validation techniques
	e.init(bin, krun->getclustering());
	
	out2 << par->kclusters << " " << e.connectivity(nn,10) << " "  << e.variance() << " " << e.dunn_max() << " " << e.silhouette() << " " << e.fmeasure(1) << " " << e.adjustedfmeasure(1) << " " << e.randindex() << " " << e.adjustedrandindex() << endl;

	
	delete krun;
    }
    }

    for (int i=0; i<par->binsize; i++) {
      delete [] neighbours[i];
    }
    for (int i=0; i<par->binsize; i++) {
      delete [] weight[i];
    }
    delete [] neighbours;
    delete [] weight;


#endif



/***************************************************************

 SOM

***************************************************************/

#ifdef SOM

    
    char name1[1000];
    char name2[1000];

    sprintf(name1, "%s.som.solution", argv[1]);
    sprintf(name2, "%s.som.control", argv[1]);


    ofstream out1(name1);
    ofstream out2(name2);
    

    int ** neighbours;
    int ** weight;
    int size = atoi(argv[2]);
    neighbours = new int * [size];
    for (int i=0; i<size; i++) {
      neighbours[i] = new int[size];
    }
    weight  = new int * [size];
    for (int i=0; i<size; i++) {
      weight[i] = new int[size];
    }


    // Range of considered number of clusters
    for (int k = 1; k <= 25; k++) {


    som * srun;

    c.evalctr = 0;


     for (c.evalctr=0; c.evalctr<c.evalnbr; c.evalctr++) {

        // Read in data 
 	databin<float> d(&c, argv[1], atoi(argv[2]), atoi(argv[3]), k);

	par = &c;
	bin = &d;


	// Run clustering algorithm
	srun = new som(&c, &d, &e);

	//	cout << "SOM Init" << endl;
	srun->init();

	c.gen_som = 10;
	c.lr0_som = 0.5;
	c.ns0_som = max(double(max(c.imax_som, c.jmax_som)/4.0),1.0);
	c.ns1_som = max(c.ns0_som / 4.0, 1.0);

	srun->run();

	c.gen_som = 40;
	c.lr0_som = 0.05;
	c.ns1_som = 1.0;

	srun->run();

	srun->classify();

	// Compute the nearest neighbor list needed for connectivity
	nnlist(neighbours, weight);

	// Apply validation techniques
	e.init(&d, srun->getclustering());

	out1 << par->kclusters << " " << e.connectivity(nn,10) << " " << e.variance() << " " << e.dunn_max() << " " << e.silhouette() << " " << e.fmeasure(1) << " " << e.adjustedfmeasure(1) << " " << e.randindex() << " " << e.adjustedrandindex() << endl;



	delete srun;

	
	par = &c;
	bin->uniformprescription();

	srun = new som(&c, &d, &e);

	srun->init();

	c.gen_som = 10;
	c.lr0_som = 0.5;
	c.ns0_som = max(double(max(c.imax_som, c.jmax_som))/4.0,1.0);
	c.ns1_som = max(c.ns0_som / 4.0, 1.0);

	srun->run();

	
	c.gen_som = 40;
	c.lr0_som = 0.05;
	c.ns1_som = 1.0;

	srun->run();

	srun->classify();


    
	// Compute the nearest neighbor list needed for connectivity
	nnlist(neighbours, weight);
    

	// Apply validation techniques
	e.init(bin, srun->getclustering());
	
	out2 << par->kclusters << " " << e.connectivity(nn,10) << " "  << e.variance() << " " << e.dunn_max() << " " << e.silhouette() << " " << e.fmeasure(1) << " " << e.adjustedfmeasure(1) << " " << e.randindex() << " " << e.adjustedrandindex() << endl;

	
	delete srun;
    }
    }

    for (int i=0; i<par->binsize; i++) {
      delete [] neighbours[i];
    }
    for (int i=0; i<par->binsize; i++) {
      delete [] weight[i];
    }
    delete [] neighbours;
    delete [] weight;


#endif


/***************************************************************

 SOTA

***************************************************************/

#ifdef SOTA

    
    char name1[1000];
    char name2[1000];

    sprintf(name1, "%s.sota.solution", argv[1]);
    sprintf(name2, "%s.sota.control", argv[1]);


    ofstream out1(name1);
    ofstream out2(name2);
    

    int ** neighbours;
    int ** weight;
    int size = atoi(argv[2]);
    neighbours = new int * [size];
    for (int i=0; i<size; i++) {
      neighbours[i] = new int[size];
    }
    weight  = new int * [size];
    for (int i=0; i<size; i++) {
      weight[i] = new int[size];
    }


    // Range of considered number of clusters
    for (int k = 1; k <= 25; k++) {


    sota * srun;

    c.evalctr = 0;


     for (c.evalctr=0; c.evalctr<c.evalnbr; c.evalctr++) {

        // Read in data 
 	databin<USED_DATA_TYPE> d(&c, argv[1], atoi(argv[2]), atoi(argv[3]), k);


	par = &c;
	bin = &d;

	srun = new sota(&c, &d, &e);
	

	srun->init();

	srun->run();
	

	srun->classify();

	// Compute the nearest neighbor list needed for connectivity
	nnlist(neighbours, weight);

	// Apply validation techniques
	e.init(&d, srun->getclustering());


	out1 << par->kclusters << " " << e.connectivity(nn,10) << " " << e.variance() << " " << e.dunn_max() << " " << e.silhouette() << " " << e.fmeasure(1) << " " << e.adjustedfmeasure(1) << " " << e.randindex() << " " << e.adjustedrandindex() << endl;


	delete srun;

	
	par = &c;
	bin->uniformprescription();

	srun = new sota(&c, &d, &e);

	srun->init();
	

	srun->run();

	
	srun->classify();


    
	// Compute the nearest neighbor list needed for connectivity
	nnlist(neighbours, weight);
    

	// Apply validation techniques
	e.init(bin, srun->getclustering());
	
	out2 << par->kclusters << " " << e.connectivity(nn,10) << " "  << e.variance() << " " << e.dunn_max() << " " << e.silhouette() << " " << e.fmeasure(1) << " " << e.adjustedfmeasure(1) << " " << e.randindex() << " " << e.adjustedrandindex() << endl;
	
	
	delete srun;
    }
    }

    for (int i=0; i<par->binsize; i++) {
      delete [] neighbours[i];
    }
    for (int i=0; i<par->binsize; i++) {
      delete [] weight[i];
    }
    delete [] neighbours;
    delete [] weight;


#endif

} 


double square(double x) {
    return x*x;
}


void fileoutput(char * name, double * values, int nbr) {
    ofstream out(name);
    for (int i=0; i<nbr; i++) {
	out << values[i] << endl;
    }
}


void output( clustering * cl, int size) {
  ofstream out("solution.dat");
  for (int i=0; i<size; i++) {
    out << (*cl)[i]+1 << endl;
  }

}

  

void nnlist(int ** nnlist, int ** weight) {
    sort(nnlist, weight);
} 
    


void sort(int ** nn_arg, int ** weight) {

    int ctr = 0;

    int size = int(0.5*par->binsize*(par->binsize-1));
    dtemp = new USED_DATA_TYPE[3*size];
 
   
    for (int i=0; i<par->binsize; i++) {
	for (int j=0; j < i; j++) {
	    dtemp[ctr] = USED_DATA_TYPE(i);
	    ctr++;
 	    dtemp[ctr] = USED_DATA_TYPE(j);
	    ctr++;
	    dtemp[ctr] = bin->precomputed_d(i,j);
	    ctr++;
	}
    }

    
    qsort((void*)dtemp, size, (3*sizeof(USED_DATA_TYPE)), lt2);
    
  

    int ptrin[par->binsize];
    for (int i=0; i<par->binsize; i++) {
	ptrin[i] = 0;
    }

  
    for (int i=0; i<size; i++) {
	int index1 = (int)dtemp[3*i];
	int index2 = (int)dtemp[3*i+1];
	nn_arg[index1][ptrin[index1]] = index2;
	weight[index1][index2] = ptrin[index1];
	ptrin[index1]++;
	nn_arg[index2][ptrin[index2]] = index1;
	weight[index2][index1] = ptrin[index2];
	ptrin[index2]++;
    }	
    nn = nn_arg;
    w = weight;

    delete [] dtemp;
}


int lt2(const void *a, const void *b) {
  USED_DATA_TYPE * pta = (USED_DATA_TYPE *)a;
  USED_DATA_TYPE * ptb = (USED_DATA_TYPE *)b;
 
    if (pta[2] < ptb[2])
	return -1;
    if (pta[2] > ptb[2])
	return 1;
 
   
    
    else return 0;
}
