/*  
    Copyright (C) 2004 Julia Handl
    Email: Julia.Handl@gmx.de

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include <iostream.h>
#include <fstream.h>
#include "databin.h"
#include "conf.h"
#include "kmeans.h"
#include "agglom.h"
#include "som.h"
#include "sota.h"
#include "time.h"

long idum;

/* Prediction strength as described by Tibshirani */

double square(double x) {
  return x*x;
}

double max(double x, double y) {
  if (x > y) return x;
  else return y;
}


int main(int argc, char ** argv) {

  idum = time(NULL);


  // Generate data set
   conf c;
   databin<float> d(&c, argv[1], atoi(argv[2]), atoi(argv[3]), atoi(argv[4]));
 

   // Initialize data structures
  float ** points = new float *[(int)(0.5*c.binsize)];
  for (int i=0; i<0.5*c.binsize; i++) {
    points[i] = new float[c.bindim];
  }

 float ** points2 = new float *[(int)(0.5*c.binsize)];
  for (int i=0; i<0.5*c.binsize; i++) {
    points2[i] = new float[c.bindim];
  }

  c.binsize = (int)(0.5*c.binsize); 

  char name1[1000];
  char name2[1000];
  

  // Open appropriate output files
#ifdef KMEANS
  sprintf(name1, "%s.kmeansnn.solution", argv[1]);
  sprintf(name2, "%s.kmeansnn.control", argv[1]);
#endif
#ifdef AGGLOM
  sprintf(name1, "%s.agglomnn.solution", argv[1]);
  sprintf(name2, "%s.agglomnn.control", argv[1]);
#endif
#ifdef SOM
  sprintf(name1, "%s.somnn.solution", argv[1]);
  sprintf(name2, "%s.somnn.control", argv[1]);
#endif
#ifdef SOTA
  sprintf(name1, "%s.sotann.solution", argv[1]);
  sprintf(name2, "%s.sotann.control", argv[1]);
#endif
#ifdef SINGLE
  sprintf(name1, "%s.singlenn.solution", argv[1]);
  sprintf(name2, "%s.singlenn.control", argv[1]);
#endif
  
  ofstream out1(name1);
  ofstream out2(name2);



  // compute stability for the original data

  // for number of clusters in the range [1, 10]
  for (int k=1; k<=10; k++) {

    cerr << k << endl;

    double pp = 0.0;

    // repeat 20 times
    for (int B=0; B<20; B++) {

      // randomly permute the data
      c.binsize = 2*c.binsize;
      d.permutate();
      c.binsize = (int)(0.5*c.binsize);

      double sum = 0.0;

      // split into two halfs

      // create first half
      for (int i=0; i<c.binsize; i++) {
	for (int j=0; j<c.bindim; j++) {
	  points[i][j] = d[i][j];
	}
      }
      
      // create second half
      for (int i=(int)(c.binsize); i<2.0*c.binsize; i++) {
	for (int j=0; j<c.bindim; j++) {
	  points2[int(i-c.binsize)][j] = d[i][j];
	}
      }
      
      // apply the clustering algorithm to the two resulting data sets
      c.kclusters = k;
      c.num_cluster = k;
      evaluation e(&c);

      databin<float> * dd = new databin<float>(&c, points);
      databin<float> * dd2 = new databin<float>(&c, points2);

#ifdef KMEANS
      kmeans * krun = new kmeans(&c, dd, &e);
      kmeans * krun2 = new kmeans(&c, dd2, &e);
#endif
#ifdef AGGLOM
      agglom * krun = new agglom(&c, dd, &e);
      agglom * krun2 = new agglom(&c, dd2, &e);
#endif
#ifdef SOM
      som * krun = new som(&c, dd, &e);
      som * krun2 = new som(&c, dd2, &e);
#endif
#ifdef SOTA
      sota * krun = new sota(&c, dd, &e);
      sota * krun2 = new sota(&c, dd2, &e);
#endif

#ifdef SINGLE
      agglom * krun = new agglom(&c, dd, &e);
      agglom * krun2 = new agglom(&c, dd2, &e);
      krun->single = true;
      krun2->single = true;
#endif
      
      krun->init();
#ifdef SOM
      c.gen_som = 10;
      c.lr0_som = 0.5;
      c.ns0_som = max(double(max(c.imax_som, c.jmax_som))/4.0,1.0);
      c.ns1_som = max(c.ns0_som / 4.0, 1.0);
#endif
      krun->run();
#ifdef SOM
      c.gen_som = 40;
      c.lr0_som = 0.05;
      c.ns1_som = 1.0;
      krun->run();
      krun->classify();
#endif
#ifdef SOTA
      krun->classify();
#endif

      krun2->init();
#ifdef SOM
      c.gen_som = 10;
      c.lr0_som = 0.5;
      c.ns0_som = max(double(max(c.imax_som, c.jmax_som))/4.0, 1.0);
      c.ns1_som = max(c.ns0_som / 4.0, 1.0);
#endif
      krun2->run();
#ifdef SOM
      c.gen_som = 40;
      c.lr0_som = 0.05;
      c.ns1_som = 1.0;
      krun2->run();
      krun2->classify();
#endif
#ifdef SOTA
      krun2->classify();
#endif

     
      

      // read in the resulting partitionings
      clustering * clust = krun->getclustering();
      clustering * clust2 = krun2->getclustering();
      

      // predict the class labels for the second half
      clustering * clust3 = new clustering(&c);
      clust3->init(k);

      // for all data items in the second half
      for (int i=(int)(c.binsize); i<2.0*c.binsize; i++) {
	double mind = 1e10;
	int minj = 0;


	// find the nearest neighbor in the first half
	for (int j=0; j<c.binsize; j++){
	  double dist = d.precomputed_d(i,j);
	  mind = e.min(mind, dist);
	  if (mind == dist) {
	    minj = j;
	  }
	}

	// assign the item to the corresponding cluster
	int k = (*clust)[minj];
	(*clust3)[int(i-c.binsize)] = k;
	

	// Alternatively, cluster centroids can be used	

	// find the closest cluster centre
	/*for (int j=0; j<k; j++) {
	  double dist = d[i].distanceto(*(new data<USED_DATA_TYPE>(&c,(*clust).centres[j])));
	  mind = e.min(mind, dist);
	  if (mind == dist) {
	    minj = j;
	  }
	  }

	// assign the item to the corresponding cluster
	(*clust3)[int(i-c.binsize)] = minj;*/
	
      }
      

      // compute Tibshirani's predictive power measure
      double r = e.tibshirani(clust2, clust3, c.binsize);

      pp += r;
      cerr << r << endl;

      delete krun;
      delete krun2;
      delete clust3;
      delete dd;
      delete dd2;

      

    }

    // output the average over 20 runs
    out1 << k << " " << pp/20.0  << endl;
  }


  // repeat process for uniformly random data
 for (int k=1; k<=10; k++) {

    double pp = 0.0;
    for (int B=0; B<20; B++) {
      c.binsize = 2*c.binsize;
      d.uniformprescription();
      c.binsize = (int)(0.5*c.binsize);

      double sum = 0.0;

      for (int i=0; i<c.binsize; i++) {
	for (int j=0; j<c.bindim; j++) {
	  points[i][j] = d[i][j];
	}
      }
      
      for (int i=(int)(c.binsize); i<2.0*c.binsize; i++) {
	for (int j=0; j<c.bindim; j++) {
	  points2[int(i-c.binsize)][j] = d[i][j];
	}
      }
      
     
      c.kclusters = k;
      c.num_cluster = k;
      evaluation e(&c);
      databin<float> * dd = new databin<float>(&c, points);
      databin<float> * dd2 = new databin<float>(&c, points2);
#ifdef KMEANS
      kmeans * krun = new kmeans(&c, dd, &e);
      kmeans * krun2 = new kmeans(&c, dd2, &e);
#endif
#ifdef AGGLOM
      agglom * krun = new agglom(&c, dd, &e);
      agglom * krun2 = new agglom(&c, dd2, &e);
#endif
#ifdef SOM
      som * krun = new som(&c, dd, &e);
      som * krun2 = new som(&c, dd2, &e);
#endif
#ifdef SOTA
      sota * krun = new sota(&c, dd, &e);
      sota * krun2 = new sota(&c, dd2, &e);
#endif
#ifdef SINGLE
      agglom * krun = new agglom(&c, dd, &e);
      agglom * krun2 = new agglom(&c, dd2, &e);
      krun->single = true;
      krun2->single = true;
#endif
    

     krun->init();
#ifdef SOM
      c.gen_som = 10;
      c.lr0_som = 0.5;
      c.ns0_som = max(double(max(c.imax_som, c.jmax_som))/4.0, 1.0);
      c.ns1_som = max(c.ns0_som / 4.0, 1.0);
#endif
      krun->run();
#ifdef SOM
      c.gen_som = 40;
      c.lr0_som = 0.05;
      c.ns1_som = 1.0;
      krun->run();
      krun->classify();
#endif
#ifdef SOTA
      krun->classify();
#endif

      krun2->init();
#ifdef SOM
      c.gen_som = 10;
      c.lr0_som = 0.5;
      c.ns0_som = max(double(max(c.imax_som, c.jmax_som))/4.0, 1.0);
      c.ns1_som = max(c.ns0_som / 4.0, 1.0);
#endif
      krun2->run();
#ifdef SOM
      c.gen_som = 40;
      c.lr0_som = 0.05;
      c.ns1_som = 1.0;
      krun2->run();
      krun2->classify();
#endif
#ifdef SOTA
      krun2->classify();
#endif

      clustering * clust = krun->getclustering();
      clustering * clust2 = krun2->getclustering();
      
     
      clustering * clust3 = new clustering(&c);
      clust3->init(k);
      
      for (int i=(int)(c.binsize); i<2.0*c.binsize; i++) {
	double mind = 1e10;
	int minj = 0;

	for (int j=0; j<k; j++) {
	  double dist = d[i].distanceto(*(new data<USED_DATA_TYPE>(&c,(*clust).centres[j])));
	  mind = e.min(mind, dist);
	  if (mind == dist) {
	    minj = j;
	  }
	}
	(*clust3)[int(i-c.binsize)] = minj;
	  
      }
      
      double r = e.tibshirani(clust2, clust3, c.binsize);

      delete krun;
      delete krun2;
      delete clust3;
      delete dd;
      delete dd2;

      pp += r;
      cerr << r << endl;
    }
    out2 << k << " " << pp/20.0  << endl;

  }

}
      
