
/*  
    Copyright (C) 2004 Julia Handl
    Email: Julia.Handl@gmx.de

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/



#include "sota.h"
#include "random.h"

extern long idum;
double lr_w = 0.01;
double lr_a = 0.005;
double lr_s = 0.001;

double sota::square(double x) {
    return x*x;
}


int * perm;


node::node(conf * par) {
    weight = new float[par->bindim];
    for (int i=0; i<par->bindim;i++) {
      weight[i] = 0.0;
    }
    var = 0.0;
    left = NULL;
    right = NULL;
    anc = NULL;
    d = 0;
    ctr = 0;
  }

/* Constructor */
sota::sota(conf * par, databin<USED_DATA_TYPE> * bin, evaluation * e):
    clalg(par, bin, e) {
  
  root = new node(par);
  

}
void del(node * root);
void del(node * root) {
  if (root->left != NULL) 
  del(root->left);
  if (root->right != NULL) 
  del(root->right);
  delete root;
}


/* Destructor */
sota::~sota() {
  if (root->left != NULL) 
    del(root->left);
  if (root->right != NULL) 
    del(root->right);
  delete root;
  
   
    
  }

/* Random intialisation */
void sota::init() {
  

  for (int i=0; i<par->binsize; i++) {
    for (int j=0; j<par->bindim; j++) {
      root->weight[j] += (*bin)[i][j];
    }
  }
  for (int j=0; j<par->bindim; j++) {
    root->weight[j] /= double(par->binsize);
    //  cerr << root->weight[j] << " ";
  }
  // cerr << endl;


  for (int i=0; i<par->binsize; i++) {
    double d = 0;
    for (int j=0; j<par->bindim; j++) {
      d += square((*bin)[i][j] - root->weight[j]);
    }
    root->var += sqrt(d);
  }
  root->var /= double(par->binsize);
  root->id = 0;
  // cerr << root->var << endl;

}

void calculate(conf * par, node * go, double * epsilon, int * id);
node * minvar(conf * par, node * x, double * var);
node * minvar(conf * par, node * x, double * var, data<USED_DATA_TYPE> * stimulus);

void calculate(conf * par, node * go, double * epsilon, int * id) {
  (*epsilon) += (go->ctr > 0)? go->d/double(go->ctr) : 0;
  go->var = (go->ctr > 0)? go->d/double(go->ctr) : 0;
  // cerr << "X " << go->ctr << " " << go->var << endl;
  /*  for (int j=0;j<par->bindim;j++) {
    cerr << go->weight[j] << " ";
  }
  cerr << endl;*/
  //  cerr << go->ctr << endl;
  go->ctr = 0;
  go->d = 0;
  if (go->right != NULL)
    calculate(par, go->left, epsilon, id);
  if (go->right != NULL)
    calculate(par, go->right, epsilon, id);

  if (go->left == NULL && go->right == NULL) {
    go->id = *id;
    (*id)++;
  }

}

node * minvar(conf * par,node * xx, double * var) {
    node * minj;
    node * next = xx;
    double minvarx = -1e10;
    double minvary = -1e10;
    double minv = -1e10;
    node * x = NULL;
    node * y = NULL;
    if (next->left != NULL) 
      x = minvar(par, next->left, &minvarx);
    if (next->right != NULL)
      y = minvar(par, next->right, &minvary);

    if (minvarx > minvary) {
      minj = x;
      minv = minvarx;
    }
    else {
      minj = y;
      minv = minvary;
    }

    if (next->left == NULL && next->right == NULL) {
      minj = next;
      minv = next->var;
    }
  
    *var = minv;
    return minj;
}


node * mindist(conf * par, node * xx, double * var, data<USED_DATA_TYPE> * stimulus  ) {
  node * minnode;
  node * current = xx;
  double mindx = 1e10;
  double mindy = 1e10;
  double mind = 1e10;
  node * x = NULL;
  node * y = NULL;
  if (current->left != NULL) 
    x = mindist(par, current->left, &mindx, stimulus);
  if (current->right != NULL)
    y = mindist(par, current->right, &mindy, stimulus);
  
  if (mindx <= mindy) {
    minnode = x;
    mind = mindx;
  }
  else {
    minnode = y;
    mind = mindy;
  }
  
  if (current->left == NULL && current->right == NULL) {
    minnode = current;
    data<USED_DATA_TYPE> * temp = new data<USED_DATA_TYPE>(par, current->weight);
    mind = stimulus->distanceto(*temp);
    delete temp;
  }

  *var = mind;
  return minnode;
}




/* Main routine of SOTA algorithm */
void sota::run() {

  //  cerr << "Clusters " << par->kclusters << endl;

 

  // increase number of clusters by one in each cycle
  for (int i=1; i<par->kclusters; i++) {
    double epsilon=-1;
    double epsilon_old=-2;


    //    cerr << i << endl;
    // find node to split (node with largest resource value)
    node * minj;
    node * next = root;
  
    double minvarx = -1e10;
    double minvary = -1e10;
    double minv = -1e10;
    node * x = NULL;
    node * y = NULL;
    if (next->left != NULL) 
      x = minvar(par, next->left, &minvarx);
    if (next->right != NULL)
      y = minvar(par, next->right, &minvary);

    if (minvarx > minvary) {
      minj = x;
      minv = minvarx;
    }
    else {
      minj = y;
      minv = minvary;
    }

    if (next->left == NULL && next->right == NULL) {
      minj = next;
      minv = next->var;
    }
    // cerr <<  minv << endl;
    
    // create two children 
    minj->left = new node(par);
    minj->right = new node(par);
    minj->left->anc = minj;
    minj->right->anc = minj;


    for (int l=0;l<par->bindim;l++) {
      minj->left->weight[l] = minj->weight[l];
      minj->right->weight[l] = minj->weight[l];
    }

    perm = new int[par->binsize];

     
 
    // run SOM
    for (int generation_ctr=1; generation_ctr <= 1000; generation_ctr++) {
      permutation();

      //      cerr << epsilon << " " << epsilon_old << " " << generation_ctr << " " << sqrt(square((epsilon-epsilon_old)/epsilon_old)) << endl;
      if (sqrt(square((epsilon-epsilon_old)/epsilon_old)) < 0.0001) break;
     

      for (int t=0; t<par->binsize; t++) {

	  data<USED_DATA_TYPE> * stimulus = &((*bin)[perm[t]]);
	
	  node * minnode;
	  node * current = root;
	  double mindx = 1e10;
	  double mindy = 1e10;
	  double mind = 1e10;
	  node * x = NULL;
	  node * y = NULL;
	  if (current->left != NULL) 
	    x = mindist(par, current->left, &mindx, stimulus);
	  if (current->right != NULL)
	    y = mindist(par, current->right, &mindy, stimulus);
	  
	  if (mindx <= mindy) {
	    minnode = x;
	    mind = mindx;
	  }
	  else {
	    minnode = y;
	    mind = mindy;
	  }

	  if (current->left == NULL && current->right == NULL) {
	    minnode = current;
	    data<USED_DATA_TYPE> * temp = new data<USED_DATA_TYPE>(par, current->weight);
	    mind = stimulus->distanceto(*temp);
	    delete temp;
	  }

	


	  // remember error
	  minnode->d += mind;
	  minnode->ctr = minnode->ctr+1;

	  // adapt weights
	  for (int l=0; l<par->bindim; l++) {
	     minnode->weight[l] = minnode->weight[l] + lr_w * ((*bin)[t][l] - minnode->weight[l] );
	  }
	  node * sister = NULL;
	  if (minnode->anc->left == minnode) {
	    sister = minnode->anc->right;
	  }
	  else {
	    sister = minnode->anc->left;
	  }

	  if (sister->left ==NULL && sister->right ==NULL) {
	    //	    cerr << "Updating sister " << endl;

	    for (int l=0; l<par->bindim; l++) {
	      sister->weight[l] = sister->weight[l] + lr_s * ((*bin)[t][l] - sister->weight[l] );
	    }
	    for (int l=0; l<par->bindim; l++) {
	      minnode->anc->weight[l] = minnode->anc->weight[l] + lr_a * ( (*bin)[t][l] - minnode->anc->weight[l]);
	    }
	  } 
	  else {
	    // cerr << "Not updating sister " << endl;
	  }
      }

      epsilon_old = epsilon;
      int id = 0;
      epsilon = 0.0;
      node * go = root;
      epsilon += (go->ctr > 0)? go->d/double(go->ctr) : 0; 
      go->var = (go->ctr > 0)? go->d/double(go->ctr) : 0;
      //      cerr << go->ctr << endl;
      go->ctr = 0;
      go->d = 0;
      if (go->left != NULL)
	calculate(par, go->left, &epsilon, &id);
      if (go->right != NULL)
      calculate(par, go->right, &epsilon, &id);
      if (go->left == NULL && go->right == NULL) {
	go->id = id;
	id++;
      }
    }
  }
}


/* SOTA, after training: determine the output neuron for 
   each data item and generate clustering object */
void sota::classify() {

    // determine output neuron for each input item
   clust = new clustering(par);
   clust->init(par->kclusters);
   data<USED_DATA_TYPE> * datacentre = new data<USED_DATA_TYPE>[par->kclusters](par);
   int * ctr = new int[par->kclusters];
   for (int i=0; i<par->kclusters; i++) {
       ctr[i] = 0;

   }


   for (int t=0; t<par->binsize; t++) {

     data<USED_DATA_TYPE> * stimulus = &((*bin)[t]);
     
	
     node * minnode;
     node * current = root;
     double mindx = 1e100;
     double mindy = 1e100;
     double mind = 1e100;
     node * x = NULL;
     node * y = NULL;
     if (current->left != NULL) 
       x = mindist(par, current->left, &mindx, stimulus);
     if (current->right != NULL)
       y = mindist(par, current->right, &mindy, stimulus);
     
     if (mindx <= mindy) {
       minnode = x;
       mind = mindx;
     }
     else {
       minnode = y;
       mind = mindy;
     }
     
     if (current->left == NULL && current->right == NULL) {
       minnode = current;
       data<USED_DATA_TYPE> * temp = new data<USED_DATA_TYPE>(par, current->weight);
       mind = stimulus->distanceto(*temp);
       delete temp;
     }
     
     // cerr << t << " " << minnode->id << endl;
     datacentre[minnode->id].add((*bin)[t]);
     (*clust)[t] = minnode->id;
     ctr[minnode->id]++;
   }
  
   for (int k=0; k<par->kclusters; k++) {
     cerr << ctr[k] << " ";
     if (ctr[k] > 0) datacentre[k].div(ctr[k]);
     clust->newcentre(k, &(datacentre[k]));
   }
   cerr << endl;

}	


// generate a permutation
void sota::permutation() {

    for (int i=0; i<par->binsize; i++) {
	perm[i] = i;
    }
    for (int i=0; i<par->binsize; i++) {
	int j = int(ran0(&idum)*(par->binsize));
	int k = int(ran0(&idum)*(par->binsize));

	int temp = perm[i];
	perm[i] = perm[k];
	perm[k] = temp;
    }
}
	  



